import _, { last, kebabCase } from 'lodash';
import path, { join, extname, relative as relative$1, removeExt, toUnix, basename, dirname } from 'upath';
import parseArgs from 'fn-args';
import path$1, { relative, sep, resolve } from 'path';
import { existsSync } from 'fs';
import staticCache from 'koa-static-cache';
import { Inject, Provide, Controller, All, ServerlessTrigger, ServerlessTriggerType } from '@midwayjs/decorator';
import { als as als$1 } from 'asynchronous-local-storage';
import { __decorate } from 'tslib';
import { init, parse } from 'es-module-lexer';
import createJITI from 'jiti';
import { sync } from 'pkg-dir';
import { inspect } from 'util';
import url from 'url';
import { sync as sync$1 } from 'globby';
import { createConfiguration } from '@midwayjs/core';
import 'axios';
export { defaults, request } from '../request/dist/midwayjs-hooks-core-request.esm.js';

function isDevelopment() {
  if (process.env.MIDWAY_TS_MODE === 'true' ||
  /* istanbul ignore next */
  process.env.NODE_ENV === 'test' ||
  /* istanbul ignore next */
  process.env.NODE_ENV === 'development') {
    return true;
  }
  /* istanbul ignore next */


  return false;
}
function useHooksMiddleware(fn) {
  return (...args) => {
    /**
     * Hooks middleware
     * const middleware = (next) => { const ctx = useContext() }
     */
    if (parseArgs(fn).length === 1) {
      const next = last(args);
      return fn(next);
    }

    return fn(...args);
  };
}
function formatCode(code) {
  try {
    const prettier = require('prettier');

    return prettier.format(code, {
      semi: true,
      singleQuote: true,
      parser: 'babel'
    });
  } catch {
    return code;
  }
}
function isPathInside(child, parent) {
  const relation = relative(parent, child);
  return Boolean(relation && relation !== '..' && !relation.startsWith(`..${sep}`) && relation !== resolve(child));
}

class FileRouter {
  constructor(options) {
    this.root = void 0;
    this.projectConfig = void 0;
    this.useSourceFile = void 0;
    const {
      root,
      projectConfig,
      useSourceFile
    } = options;
    this.root = root;
    this.projectConfig = projectConfig;
    this.useSourceFile = useSourceFile;
  }

  get source() {
    if (this.useSourceFile) {
      return join(this.root, this.projectConfig.source);
    }

    return join(this.root, this.projectConfig.build.outDir);
  } // src/apis/lambda


  getApiDirectory(baseDir) {
    return join(this.source, baseDir);
  }

  getRoute(file) {
    const {
      routes
    } = this.projectConfig;
    return routes.find(route => {
      const dir = this.getApiDirectory(route.baseDir);
      return this.inside(file, dir);
    });
  }

  getGatewayByRoute(route) {
    const gateway = this.projectConfig.gateway.find(adapter => adapter.is(route));

    if (!gateway) {
      throw new Error(`Can't find the correct gateway adapter, please check if midway.config.ts is correct`);
    }

    return gateway;
  }

  isApiFile(file) {
    if (file.endsWith('.test.ts') || file.endsWith('.test.js')) {
      return false;
    }

    if (extname(file) !== '.ts' && extname(file) !== '.js') {
      return false;
    }

    const route = this.getRoute(file);
    return !!route;
  }

  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = relative$1(this.source, file); // src/apis/lambda/index.ts -> apis-lambda-index

    const id = kebabCase(removeExt(relativePath, extname(relativePath)));
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('');
    return name.toLowerCase();
  }

  inside(child, parent) {
    return isPathInside(toUnix(child), toUnix(parent));
  }

}

/**
 * @private
 * private api, may change without notice.
 * Use asynchronous-local-storage due to serverless environment does not support node.js 12.17.0
 */
const als = {
  getStore(key) {
    return als$1.get(key);
  },

  run(ctx, callback) {
    return new Promise(resolve => {
      als$1.runWith(async () => {
        resolve(await callback());
      }, ctx);
    });
  }

};
function useContext() {
  return als.getStore('ctx');
}
function createFunctionContainer(options) {
  const {
    fn,
    functionId,
    parseArgs,
    handlerDecorators = [],
    classDecorators = [],
    isHTTP
  } = options;
  let FunctionContainer = class FunctionContainer {
    constructor() {
      this.ctx = void 0;
    }

    async handler(...handlerArgs) {
      const args = parseArgs(this.ctx, ...handlerArgs);
      return isHTTP ? await fn(...args) : await als.run({
        ctx: this.ctx
      }, async () => await fn(...args));
    }

  };

  __decorate([Inject()], FunctionContainer.prototype, 'ctx', void 0);

  __decorate([...handlerDecorators], FunctionContainer.prototype, 'handler', null);

  FunctionContainer = __decorate([Provide(functionId), ...classDecorators], FunctionContainer);
  return FunctionContainer;
}

async function createHTTPApiClient(file, code, router) {
  await init;
  const [, exports] = parse(code);

  if (exports.length === 0) {
    return null;
  }

  const baseUrl = router.getBaseUrl(file);
  const {
    request: {
      client
    }
  } = router.projectConfig;
  const funcs = exports.map(name => {
    const isExportDefault = name === 'default';
    return {
      url: getUrl(baseUrl, name),
      meta: {},
      functionId: router.getFunctionId(file, name, isExportDefault),
      isExportDefault: isExportDefault,
      name: isExportDefault ? '$default' : name
    };
  });
  return formatCode(buildClient$1(client, funcs));
}

function buildClient$1(client, funcs) {
  const buildFunction = func => {
    return `
      export ${func.isExportDefault ? 'default' : ''} function ${func.name} (...args) {
        return request({
          url: '${func.url}',
          method: args.length === 0 ? 'GET' : 'POST',
          data: { args },
          meta: ${JSON.stringify(func.meta, null, 2)}
        })
      }
    `;
  };

  const functions = funcs.map(buildFunction).join('\n');
  return `
    import { request } from '${client}';
    ${functions}
  `;
}

function getUrl(baseUrl, name) {
  if (name === 'default') {
    return baseUrl;
  }

  if (baseUrl.endsWith('/')) {
    return `${baseUrl}${name}`;
  }

  return `${baseUrl}/${name}`;
}

function duplicateLogger(root, existPath, currentPath, api) {
  console.log('[%s] Duplicate routes detected. %s and %s both resolve to %s.', 'warn', relative$1(root, existPath), relative$1(root, currentPath), api);
}

class HTTPRouter extends FileRouter {
  constructor(...args) {
    super(...args);
    this.routes = new Map();
  }

  getBaseUrl(file) {
    const url = this.getHTTPPath(file, '', true);

    if (url === '/*') {
      return '/';
    }

    if (url.endsWith('/*')) {
      return url.slice(0, -2);
    }

    return url;
  }

  getHTTPPath(file, method, isExportDefault) {
    const {
      basePath,
      baseDir,
      underscore
    } = this.getRoute(file);
    const lambdaDirectory = this.getApiDirectory(baseDir);
    const {
      isCatchAllRoutes,
      filename
    } = this.parseFilename(basename(file, extname(file)));
    const fileRoute = filename === 'index' ? '' : filename; // TODO remove underscore support in future

    const func = isExportDefault ? '' : `${underscore ? '_' : ''}${method}`;
    const api = toUnix(join(basePath,
    /**
     * /apis/lambda/index.ts -> ''
     * /apis/lambda/note/index.ts -> 'note'
     */
    relative$1(lambdaDirectory, dirname(file)),
    /**
     * index -> ''
     * demo -> '/demo'
     */
    fileRoute, // underscore: getNoteList -> _getNoteList
    func, isCatchAllRoutes ? '/*' : ''));
    /**
     * duplicate routes
     */

    const existPath = this.routes.get(api);

    if (existPath && existPath !== toUnix(file)) {
      duplicateLogger(this.root, existPath, file, api);
    }

    this.routes.set(api, toUnix(file));
    return api;
  }

  parseFilename(filename) {
    var _re$exec;

    const re = /\[\.{3}(.+)]/;
    const isCatchAllRoutes = re.test(filename);
    return {
      isCatchAllRoutes,
      filename: isCatchAllRoutes ? (_re$exec = re.exec(filename)) === null || _re$exec === void 0 ? void 0 : _re$exec[1] : filename
    };
  }

}

class HTTPGateway {
  static is(route) {
    return !!(route !== null && route !== void 0 && route.basePath);
  }

  constructor(options) {
    this.options = void 0;
    this.container = void 0;
    this.router = void 0;
    this.options = options;
    this.router = new HTTPRouter({
      root: this.options.root,
      projectConfig: this.options.projectConfig,
      useSourceFile: this.options.router.useSourceFile
    });
  }

  createApi(options) {
    const {
      file,
      functionName,
      isExportDefault,
      fn
    } = options;
    const httpPath = this.router.getHTTPPath(file, functionName, isExportDefault);
    fn._param.url = httpPath;
    this.createHTTPApi(httpPath, options);
  }

  createHTTPApi(httpPath, options) {
    const {
      functionId,
      fn
    } = options; // setup middleware

    const middleware = [...fn.middleware];
    const FunctionContainer = createFunctionContainer({
      isHTTP: true,
      fn,
      functionId,

      parseArgs(ctx) {
        var _ctx$request, _ctx$request$body;

        const args = ((_ctx$request = ctx.request) === null || _ctx$request === void 0 ? void 0 : (_ctx$request$body = _ctx$request.body) === null || _ctx$request$body === void 0 ? void 0 : _ctx$request$body.args) || [];
        return args;
      },

      classDecorators: [Controller('/')],
      handlerDecorators: [All(httpPath, {
        middleware
      })]
    });
    this.container.bind(functionId, FunctionContainer);
  }

  afterCreate() {
    if (isDevelopment()) {
      return;
    }

    if (!this.isViteProject) {
      return;
    }

    const {
      routes
    } = this.router;

    if (routes.has('/') || routes.has('/*')) {
      return;
    }

    const baseDir = this.container.get('baseDir');
    const mw = staticCache({
      dir: join(baseDir, '..', this.options.projectConfig.build.viteOutDir),
      dynamic: true,
      alias: {
        '/': 'index.html',

        /**
         * Add alias for windows, '/' -> '\\'
         * https://github.com/koajs/static-cache/blob/master/index.js#L45
         */
        '\\': 'index.html'
      },
      buffer: true,
      gzip: true
    });

    const fn = async () => {};

    fn.middleware = [mw];
    this.createHTTPApi('/*', {
      functionId: 'hooks-host',
      fn
    });
  }

  get isViteProject() {
    return ['vite.config.ts', 'vite.config.js'].some(config => existsSync(join(this.options.root, config)));
  }

}
HTTPGateway.router = HTTPRouter;
HTTPGateway.createApiClient = createHTTPApiClient;

async function createEventApiClient(file, code, router) {
  await init;
  const [, exports] = parse(code);

  if (exports.length === 0) {
    return null;
  }

  const {
    request: {
      client
    }
  } = router.projectConfig;
  const funcs = exports.map(name => {
    const isExportDefault = name === 'default';
    return {
      functionId: router.getFunctionId(file, name, isExportDefault),
      isExportDefault: isExportDefault,
      name: isExportDefault ? '$default' : name
    };
  });
  return formatCode(buildClient(client, funcs));
}

function buildClient(client, funcs) {
  const buildFunction = func => {
    return `
      export ${func.isExportDefault ? 'default' : ''} function ${func.name} (...args) {
        return request({
          functionId: '${func.functionId}',
          data: { args }
        })
      }
    `;
  };

  const functions = funcs.map(buildFunction).join('\n');
  return `
    import { request } from '${client}';
    ${functions}
  `;
}

class EventGateway {
  constructor() {
    this.container = void 0;
  }

  static is(route) {
    return !!(route !== null && route !== void 0 && route.event);
  }

  createApi(options) {
    const {
      functionId,
      fn,
      route
    } = options;
    const FunctionContainer = createFunctionContainer({
      fn,
      functionId,

      parseArgs(_, event) {
        let args;

        switch (route.event) {
          case 'wechat-miniprogram':
            args = event === null || event === void 0 ? void 0 : event.args;
        }

        return args || [];
      },

      handlerDecorators: [ServerlessTrigger(ServerlessTriggerType.EVENT, {
        functionName: functionId
      })]
    });
    this.container.bind(functionId, FunctionContainer);
  }

}
EventGateway.router = FileRouter;
EventGateway.createApiClient = createEventApiClient;

//   constructor(actual: any, name: string, reason = 'is invalid') {
//     const message = `[ERR_INVALID_ARG_VALUE]: The argument '${name}' ${reason} ${getTypeMessage(
//       actual
//     )}`
//     super(message)
//   }
// }

class ERR_INVALID_ARG_TYPE extends Error {
  constructor(name, expected, actual) {
    const message = `[ERR_INVALID_ARG_TYPE]: The '${name}' argument must be of type ${expected}${getTypeMessage(actual)}`;
    super(message);
  }

} // fork from https://github.com/nodejs/node/blob/master/lib/internal/errors.js

function getTypeMessage(actual) {
  let msg = '';

  if (actual == null) {
    msg += `. Received ${actual}`;
  } else if (typeof actual === 'function' && actual.name) {
    msg += `. Received function ${actual.name}`;
  } else if (typeof actual === 'object') {
    if (actual.constructor && actual.constructor.name) {
      msg += `. Received an instance of ${actual.constructor.name}`;
    } else {
      const inspected = inspect(actual, {
        depth: -1
      });
      msg += `. Received ${inspected}`;
    }
  } else {
    let inspected = inspect(actual, {
      colors: false
    });
    if (inspected.length > 25) inspected = `${inspected.slice(0, 25)}...`;
    msg += `. Received type ${typeof actual} (${inspected})`;
  }

  return msg;
}

function validateArray(value, name) {
  if (!Array.isArray(value)) {
    throw new ERR_INVALID_ARG_TYPE(name, 'Array', value);
  }
}
function validateString(value, name) {
  if (typeof value !== 'string') throw new ERR_INVALID_ARG_TYPE(name, 'string', value);
} // export function validateOneOf(value: any, name: string, oneOf: any[]) {
//   if (!oneOf.includes(value)) {
//     const allowed = oneOf
//       .map((v) => (typeof v === 'string' ? `'${v}'` : String(v)))
//       .join(', ')
//     const reason = 'must be one of: ' + allowed
//     throw new ERR_INVALID_ARG_VALUE(name, value, reason)
//   }
// }

function validateFunction(value, name) {
  if (typeof value !== 'function') {
    throw new ERR_INVALID_ARG_TYPE(name, 'Function', value);
  }
}

// Ignore Vite dev server
const VITE_REQUEST = ['@vite', '@react-refresh']; // Ref: https://github.com/vitejs/vite/blob/d156a9f364dddcfc41338f83e39db38a00a2ceb0/packages/vite/src/node/constants.ts#L49

const KNOWN_ASSET_TYPES = [// images
'png', 'jpe?g', 'gif', 'svg', 'ico', 'webp', 'avif', // media
'mp4', 'webm', 'ogg', 'mp3', 'wav', 'flac', 'aac', // fonts
'woff2?', 'eot', 'ttf', 'otf', // css
'css', 'less', 'sass', 'scss', // script
'm?jsx?', 'tsx?', 'json', 'map', // other
'wasm'];
const DEFAULT_ASSETS_RE = new RegExp(`\\.(` + KNOWN_ASSET_TYPES.join('|') + `)(\\?.*)?$`);
const ignorePattern = req => {
  if (VITE_REQUEST.some(api => req.url.includes(api))) {
    return true;
  }

  const {
    pathname,
    query
  } = url.parse(req.url);
  return DEFAULT_ASSETS_RE.test(pathname) || DEFAULT_ASSETS_RE.test(query);
};

function getProjectRoot(cwd) {
  return sync(cwd) || process.cwd();
}
function getConfig(cwd) {
  const root = getProjectRoot(cwd);
  const configs = {
    ts: path.join(root, 'midway.config.ts'),
    js: path.join(root, 'midway.config.js')
  };

  if (!existsSync(configs.ts) && !existsSync(configs.js)) {
    throw new Error(`[ERR_INVALID_CONFIG] midway.config.ts is not found, root: ${root}`);
  }

  const userConfig = existsSync(configs.ts) ? requireByJiti(configs.ts) : requireByJiti(configs.js);
  const builtinGateways = new Set();

  for (const route of userConfig.routes) {
    if (route.basePath) builtinGateways.add(HTTPGateway);
    if (route.event) builtinGateways.add(EventGateway);
  }

  userConfig.gateway = userConfig.gateway || [];
  userConfig.gateway.push(...builtinGateways);
  return _.defaultsDeep({}, userConfig, {
    source: './src/apis',
    dev: {
      ignorePattern
    },
    build: {
      viteOutDir: './build',
      outDir: './dist'
    },
    request: {
      client: '@midwayjs/hooks-core/request'
    }
  });
}
function defineConfig(config) {
  validateArray(config.routes, 'defineConfig.routes');
  return config;
}

const requireByJiti = id => {
  const jiti = createJITI();
  const contents = jiti(id);
  if ('default' in contents) return contents.default;
  return contents;
};

class HooksComponent {
  constructor(options) {
    this.options = void 0;
    this.adapters = void 0;

    this.useAsyncLocalStorage = async (ctx, next) => {
      await als.run({
        ctx
      }, async () => {
        await next();
      });
    };

    this.options = options;
    this.adapters = options.projectConfig.gateway.map(Adapter => new Adapter(this.options));
  }

  init() {
    const {
      router,
      projectConfig: {
        routes
      }
    } = this.options;
    let count = 0;
    const totalCount = routes.reduce((totalCount, route) => {
      // Windows
      const dir = path.join(router.source, route.baseDir);
      const files = sync$1([dir]).filter(file => router.isApiFile(file));
      return totalCount + files.length;
    }, 0);
    const configuration = createConfiguration({
      namespace: '@midwayjs/hooks',
      directoryResolveFilter: routes.map(route => {
        return {
          // Windows
          pattern: path$1.join(router.source, route.baseDir),
          ignoreRequire: true,
          filter: (_, file, container) => {
            if (!router.isApiFile(file)) {
              return;
            } // Initialize container


            this.adapters.forEach(adapter => adapter.container = container); // Create api function

            const adapter = this.getAdapterByRoute(route);

            const mod = require(file);

            this.createApi(mod, adapter, file, route);
            container.bindClass(mod, '', file); // Call afterCreate hooks

            count++;

            if (count === totalCount) {
              this.adapters.forEach(adapter => {
                var _adapter$afterCreate;

                return (_adapter$afterCreate = adapter.afterCreate) === null || _adapter$afterCreate === void 0 ? void 0 : _adapter$afterCreate.call(adapter);
              });
            }
          }
        };
      })
    });
    configuration.onReady((_, app) => {
      // Setup global middleware
      for (const mw of this.getGlobalMiddleware()) {
        app.use(mw);
      }
    }).onStop(_.noop);
    return {
      Configuration: configuration
    };
  }

  getAdapterByRoute(route) {
    const Adapter = this.options.router.getGatewayByRoute(route);
    return this.adapters.find(inst => inst instanceof Adapter);
  }

  createApi(mod, adapter, file, route) {
    var _mod$config;

    const modMiddleware = (mod === null || mod === void 0 ? void 0 : (_mod$config = mod.config) === null || _mod$config === void 0 ? void 0 : _mod$config.middleware) || [];

    const funcs = _.pickBy(mod, _.isFunction);

    for (const [name, fn] of Object.entries(funcs)) {
      fn.middleware = (fn.middleware || (fn.middleware = [])).concat(modMiddleware).map(useHooksMiddleware);
      const isExportDefault = name === 'default';
      const functionName = isExportDefault ? '$default' : name;
      const functionId = this.options.router.getFunctionId(file, functionName, isExportDefault);
      fn._param = {
        functionId
      };
      adapter.createApi({
        fn,
        functionName,
        isExportDefault,
        functionId,
        file,
        route
      });
    }
  }
  /**
   * @description Web only
   */


  getGlobalMiddleware() {
    var _this$options$runtime, _this$options$runtime2;

    const mws = [this.useAsyncLocalStorage];
    (_this$options$runtime = this.options.runtimeConfig.middleware) === null || _this$options$runtime === void 0 ? void 0 : (_this$options$runtime2 = _this$options$runtime.forEach) === null || _this$options$runtime2 === void 0 ? void 0 : _this$options$runtime2.call(_this$options$runtime, mw => mws.push(useHooksMiddleware(mw)));
    return mws;
  }

}

/**
 * @description Create hooks component
 */

const hooks = (runtimeConfig = {}) => {
  if (runtimeConfig.middleware !== undefined) {
    validateArray(runtimeConfig.middleware, 'runtimeConfig.middleware');
  }

  const root = getProjectRoot();
  const projectConfig = getConfig();
  const router = new FileRouter({
    root,
    projectConfig,
    useSourceFile: isDevelopment()
  });
  const hooks = new HooksComponent({
    root,
    router,
    runtimeConfig,
    projectConfig
  });
  return hooks.init();
};

function useLogger() {
  const ctx = useContext();
  return ctx.logger;
}
function usePlugin(key) {
  validateString(key, 'key');
  const ctx = useContext();
  return ctx.app[key] || ctx[key];
}
async function useInject(identifier) {
  if (typeof identifier !== 'function' && typeof identifier !== 'string') {
    throw new ERR_INVALID_ARG_TYPE('identifier', 'class, string', identifier);
  }

  const ctx = useContext();
  const requestContext = ctx['requestContext'];
  return requestContext.getAsync(identifier);
}
function useConfig(key) {
  if (key !== undefined) {
    validateString(key, 'key');
  }

  const ctx = useContext();
  const requestContext = ctx['requestContext'];
  return requestContext.getConfigService().getConfiguration(key);
}

export { ERR_INVALID_ARG_TYPE, EventGateway, FileRouter, HTTPGateway, HTTPRouter, als, createEventApiClient, createFunctionContainer, createHTTPApiClient, defineConfig, formatCode, getConfig, getProjectRoot, hooks, isDevelopment, isPathInside, useConfig, useContext, useHooksMiddleware, useInject, useLogger, usePlugin, validateArray, validateFunction, validateString };
